#include <linux/list.h>
#include <linux/semaphore.h>

#include "../common.h"
#include "../find.c"
#include "../CR0.c"

typedef struct {
   struct list_head link;
   policy_mess_t mbuf;
} policy_context_t;

static struct semaphore req_wait;                               // length request queue
LIST_HEAD( req_queue );                                         // request queue 

static unsigned int lenblk( void ) {                            // current request queue length
   unsigned int num = 0;
   struct list_head *iter;
   if( list_empty( &req_queue ) ) return 0;
   list_for_each( iter, &req_queue ) num++;
   return num;
}

static uint32_t sequence = 0;

static policy_context_t* addblk( syscall_type_t type, struct sockaddr_in *usin ) {                           // add new request to queue
   policy_context_t *elem = (policy_context_t*)vmalloc( sizeof( policy_context_t ) );
   if( elem != NULL ) {
      list_add_tail( &elem->link, &req_queue );
      elem->mbuf.type = type;
      elem->mbuf.sequence = ++sequence;
      elem->mbuf.ipn = usin->sin_addr.s_addr;
      elem->mbuf.portn = usin->sin_port;
      sprintf( elem->mbuf.text, "[%d] %s_address=%s %s_port=%d", lenblk(), 
               ( ACCEPT_BEFORE == type ? "local" : "remote" ), in4_ntoa( ntohl( usin->sin_addr.s_addr ) ),
               ( ACCEPT_BEFORE == type ? "local" : "remote" ), ntohs( usin->sin_port ) );
      DEB( "allocate request block %p sequence=%u\n", elem, elem->mbuf.sequence );
      up( &req_wait );
   }
   return elem;
}

static void erase( void ) {                                            // delete head request
   policy_context_t *head;
   if( list_empty( &req_queue ) ) return;
   head = list_entry( req_queue.next, policy_context_t, link );
   DEB( "erase request block %p sequence=%u\n", head, head->mbuf.sequence );
   list_del( req_queue.next );
   vfree( head );
}

// <linux/syscalls.h> :
asmlinkage long (*old_sys_socketcall) ( int call, unsigned long __user *args );

static int already = 0;                        // open device flag - filtering on

#define PARMS 3
long get_addr( struct sockaddr *sa, unsigned long __user *args ) {          // restore syscall sockaddr  
   const unsigned int len = PARMS * sizeof( unsigned long );
   unsigned long a[ PARMS ];
   if( copy_from_user( a, args, len ) )
      return -EFAULT;
   if( copy_from_user( sa, (struct sockaddr __user*)a[ 1 ], sizeof( struct sockaddr ) ) )
      return -EFAULT;
   return 0;
}

/* <linux/socket.h> :
      struct sockaddr {
         sa_family_t     sa_family;            // address family, AF_xxx
         char            sa_data[14];          // 14 bytes of protocol address
   }; */
/* <linux/in.h> :
      struct sockaddr_in {
         __kernel_sa_family_t  sin_family;     // Address family
         __be16                sin_port;       // Port number
         struct in_addr        sin_addr;       // Internet address
            ... 
      };
      struct in_addr {
         __be32  s_addr;
      }; */
/* <linux/byteorder/generic.h> :
      #define ___ntohl(x) __be32_to_cpu(x)
      #define ___ntohs(x) __be16_to_cpu(x)
      #define ntohl(x) ___ntohl(x)
      #define ntohs(x) ___ntohs(x) */

asmlinkage long new_sys_socketcall( int call, unsigned long __user *args ) {
   long ret;
   if( already && SYS_ACCEPT == call ) {                  // accept() before syscall
      long err;
      struct sockaddr sa;
      if( ( err = get_addr( &sa, args ) ) < 0 ) return err;
      if( AF_INET == sa.sa_family ) {                       // only IPv4
         policy_context_t *pc = addblk( ACCEPT_BEFORE, (struct sockaddr_in *)&sa );
         if( NULL == pc ) return -ENOMEM;
         DEB( "%s\n", pc->mbuf.text  );
      }
   } 
   if( already && SYS_CONNECT == call ) {                              // connect() before syscall
      long err;
      struct sockaddr sa;
      if( ( err = get_addr( &sa, args ) ) < 0 ) return err;
      if( AF_INET == sa.sa_family ) {                       // only IPv4
         policy_context_t *pc = addblk( CONNECT, (struct sockaddr_in *)&sa );
         if( NULL == pc ) return -ENOMEM;
         DEB( "%s\n", pc->mbuf.text  );
      }
   } 
   ret = old_sys_socketcall( call, args );                  // retranslate to original sys_socketcall()
   if( already && SYS_ACCEPT == call ) {                               // accepr() after syscall
      long err;
      struct sockaddr sa;
      if( ( err = get_addr( &sa, args ) ) < 0 ) return err;
      if( AF_INET == sa.sa_family ) {                       // only IPv4
         policy_context_t *pc = addblk( ACCEPT_AFTER, (struct sockaddr_in *)&sa );
         if( NULL == pc ) return -ENOMEM;
         DEB( "%s\n", pc->mbuf.text  );
      } 
   }
   return ret;
}
EXPORT_SYMBOL( new_sys_socketcall );

static ssize_t read( struct file *file, char *buf, size_t count, loff_t *ppos ) {
   int len = count < sizeof( policy_mess_t ) ? count : sizeof( policy_mess_t ); 
   policy_context_t *head;
   int ret = down_killable( &req_wait );
   if( ret != 0 ) return ret;
   head = list_entry( req_queue.next, policy_context_t, link );
   if( copy_to_user( buf, &head->mbuf, len ) ) return -EFAULT;
   erase();
   if( list_empty( &req_queue ) ) 
      ret = down_trylock( &req_wait );
   return len;
}

static ssize_t write( struct file *file, const char *buf, size_t count, loff_t *ppos ) {
   policy_perm_t perm;
   int ret = copy_from_user( &perm, (void*)buf, sizeof( policy_perm_t ) );
   if( ret != 0 ) return ret;
   DEB( "write: sequence %u %s\n", perm.sequence, ( ALLOW == perm.type ? "allow" : "deny" ) );
   return sizeof( policy_perm_t );
}

static int open( struct inode *n, struct file *f ) {
   if( already ) {                                      // there may not be duplicate devices
      ERR( "device /dev/%s is busy\n", DEVNAME );
      return -EBUSY;
   }
   already++;
   DEB( "filtering set on\n" );
   sema_init( &req_wait, 0 );
   return 0;
}

static int release( struct inode *n, struct file *f ) {
   already--;
   DEB( "filtering set off\n" );
   while( !list_empty( &req_queue ) ) erase();
   return 0;
}

static const struct file_operations fw_fops = {
   .owner   = THIS_MODULE,
   .open    = open,
   .release = release,
   .read    = read,
   .write   = write,
};

static struct miscdevice fw_dev = { MISC_DYNAMIC_MINOR, DEVNAME, &fw_fops };

static void **taddr;                                        // table sys_call_table address

static int __init init( void ) {
   void *waddr;
   // ----------- nested functions are a GCC extension ---------
   int pos_in_table( const char *symbol ) {                 // position in sys_call_table (__NR_*)
      const int last = __NR_process_vm_writev;              // near last syscall in i386
      int n;
      waddr = find_sym( symbol );
      if( NULL == waddr ) return -1;
      for( n = 0; n <= last; n++ )
         if( taddr[ n ] == waddr ) break;
      return n <= last ? n : -1;
   }
   // --------------------------------------------------------
   void show_in_table( char *symb ) {                       // print info about symbol
      waddr = find_sym( symb );
      if( NULL == waddr ) {
         DEB( "symbol %s not found in kernel\n", symb );
      }
      else {
         int n = pos_in_table( symb );
         if( n > 0 ) {
            DEB( "symbol %s address = %p, position in sys_call_table = %d\n", symb, waddr, n );
         }
         else
            DEB( "symbol %s address = %p, not found in sys_call_table\n", symb, waddr );
      }
   }
   // --------------------------------------------------------
   int ret = misc_register( &fw_dev );                      // register /dev/policy
   if( ret ) {
      ERR( "unable to register device /dev/%s\n", DEVNAME );
      return ret;
   }
   LOG( "registered device /dev/%s\n", DEVNAME );
   if( NULL == ( taddr = find_sym( "sys_call_table" ) ) ) {
      ERR( "sys_call_table not found\n" ); 
      goto err;
   }
   DEB( "sys_call_table address = %p\n", taddr );
   old_sys_socketcall = (void*)taddr[ __NR_socketcall ];
   if( NULL == ( waddr = find_sym( "sys_socketcall" ) ) ) { // sys_socketcall not exported
      ERR( "sys_socketcall not found\n" ); 
      goto err;
   }
   if( old_sys_socketcall != waddr ) {                      // reinsurance!
      ERR( "Oooops! I don't understand: addresses not equal\n" );
      goto err;
   }
   if( debug ) show_cr0();
   rw_enable();
   taddr[ __NR_socketcall ] = new_sys_socketcall;
   if( debug ) show_cr0();
   rw_disable();
   if( debug ) show_cr0();
   LOG( "install new sys_socketcall handler: %p\n", &new_sys_socketcall );
   return 0;
err:
   misc_deregister( &fw_dev );                              // deregister /dev/policy
   return -EINVAL;
}

static void __exit exit( void ) {
   LOG( "sys_socketcall handler before unload: %p\n", (void*)taddr[ __NR_socketcall ] );
   rw_enable();
   taddr[ __NR_socketcall ] = old_sys_socketcall;
   rw_disable();
   LOG( "restore old sys_socketcall handler: %p\n", (void*)taddr[ __NR_socketcall ] );
   misc_deregister( &fw_dev );                              // deregister /dev/policy
   LOG( "deregistered device /dev/%s\n", DEVNAME );
   return;
}

module_init( init );
module_exit( exit );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Oleg Tsiliuric <olej.tsil@gmail.com>" );
MODULE_VERSION( "2.4" );
